#include <limits.h>

#include "t3f/t3f.h"

typedef struct FLOODED_LINE      /* store segments which have been flooded */
{
   short flags;                  /* status of the segment */
   short lpos, rpos;             /* left and right ends of segment */
   short y;                      /* y coordinate of the segment */
   int next;                     /* linked list if several per line */
} FLOODED_LINE;

/* Note: a 'short' is not sufficient for 'next' above in some corner cases. */


static int flood_count;          /* number of flooded segments */
static void * flood_scratch = NULL;
static int flood_scratch_size = 0;

#define FLOOD_IN_USE             1
#define FLOOD_TODO_ABOVE         2
#define FLOOD_TODO_BELOW         4

#define FLOOD_LINE(c)            (((FLOODED_LINE *)flood_scratch) + c)

static bool same_color(ALLEGRO_COLOR c1, ALLEGRO_COLOR c2)
{
    unsigned char r[2], g[2], b[2], a[2];

    al_unmap_rgba(c1, &r[0], &g[0], &b[0], &a[0]);
    al_unmap_rgba(c2, &r[1], &g[1], &b[1], &a[1]);
    if(r[0] != r[1] || g[0] != g[1] || b[1] != b[1] || a[0] !=a[1])
    {
        return false;
    }

    return true;
}

static void _grow_scratch_mem(int size)
{
    if(size > 0)
    {
        flood_scratch = al_realloc(flood_scratch, flood_scratch_size + size);
        flood_scratch_size += size;
    }
    else
    {
        flood_scratch = al_malloc(size);
        flood_scratch_size = size;
    }
}

static void _reset_scratch_mem(void)
{
    if(flood_scratch)
    {
        free(flood_scratch);
        flood_scratch = NULL;
    }
    flood_scratch_size = 0;
}

/* flooder:
 *  Fills a horizontal line around the specified position, and adds it
 *  to the list of drawn segments. Returns the first x coordinate after
 *  the part of the line which it has dealt with.
 */
static int flooder(ALLEGRO_BITMAP *bmp, int x, int y, ALLEGRO_COLOR src_color, ALLEGRO_COLOR dest_color)
{
   FLOODED_LINE *p;
   int left = 0, right = 0;
   int c, i;

      /* check start pixel */
      if (!same_color(al_get_pixel(bmp, x, y), src_color))
	 return x+1;

      /* work left from starting point */
      for (left=x-1; left>=0; left--)
	 if (!same_color(al_get_pixel(bmp, left, y), src_color))
	    break;

      /* work right from starting point */
      for (right=x+1; right<al_get_bitmap_width(bmp); right++)
	 if (!same_color(al_get_pixel(bmp, right, y), src_color))
	    break;

   left++;
   right--;

   /* draw the line */
   for(i = left; i <= right; i++)
   {
       al_put_pixel(i, y, dest_color);
   }

   /* store it in the list of flooded segments */
   c = y;
   p = FLOOD_LINE(c);

   if (p->flags) {
      while (p->next) {
	 c = p->next;
	 p = FLOOD_LINE(c);
      }

      p->next = c = flood_count++;
      _grow_scratch_mem(sizeof(FLOODED_LINE) * flood_count);
      p = FLOOD_LINE(c);
   }

   p->flags = FLOOD_IN_USE;
   p->lpos = left;
   p->rpos = right;
   p->y = y;
   p->next = 0;

   if (y > 0)
      p->flags |= FLOOD_TODO_ABOVE;

   if (y+1 < al_get_bitmap_height(bmp))
      p->flags |= FLOOD_TODO_BELOW;

   return right+2;
}



/* check_flood_line:
 *  Checks a line segment, using the scratch buffer is to store a list of
 *  segments which have already been drawn in order to minimise the required
 *  number of tests.
 */
static int check_flood_line(ALLEGRO_BITMAP *bmp, int y, int left, int right, ALLEGRO_COLOR src_color, ALLEGRO_COLOR dest_color)
{
   int c;
   FLOODED_LINE *p;
   bool ret = false;

   while (left <= right) {
      c = y;

      for (;;) {
	 p = FLOOD_LINE(c);

	 if ((left >= p->lpos) && (left <= p->rpos)) {
	    left = p->rpos+2;
	    break;
	 }

	 c = p->next;

	 if (!c) {
	    left = flooder(bmp, left, y, src_color, dest_color);
	    ret = true;
	    break;
	 }
      }
   }

   return ret;
}



/* floodfill:
 *  Fills an enclosed area (starting at point x, y) with the specified color.
 */
void flood_fill(ALLEGRO_BITMAP *bmp, int x, int y, ALLEGRO_COLOR color)
{
   ALLEGRO_COLOR src_color;
   int c;
   bool done;
   FLOODED_LINE *p;
   ALLEGRO_STATE old_state;

   /* make sure we have a valid starting point */
//   if ((x < bmp->cl) || (x >= bmp->cr) || (y < bmp->ct) || (y >= bmp->cb))
//      return;

    al_store_state(&old_state, ALLEGRO_STATE_TARGET_BITMAP);
    al_set_target_bitmap(bmp);
    al_lock_bitmap(bmp, al_get_bitmap_format(bmp), ALLEGRO_LOCK_READWRITE);

   /* what color to replace? */
   src_color = al_get_pixel(bmp, x, y);
   if (same_color(src_color, color)) {
       al_unlock_bitmap(bmp);
       al_restore_state(&old_state);
      return;
   }

   /* set up the list of flooded segments */
   _reset_scratch_mem();
   _grow_scratch_mem(sizeof(FLOODED_LINE) * al_get_bitmap_height(bmp));
   flood_count = al_get_bitmap_height(bmp);
   p = flood_scratch;
   for (c=0; c<flood_count; c++) {
      p[c].flags = 0;
      p[c].lpos = SHRT_MAX;
      p[c].rpos = SHRT_MIN;
      p[c].y = y;
      p[c].next = 0;
   }

   /* start up the flood algorithm */
   flooder(bmp, x, y, src_color, color);

   /* continue as long as there are some segments still to test */
   do {
      done = true;

      /* for each line on the screen */
      for (c=0; c<flood_count; c++) {

	 p = FLOOD_LINE(c);

	 /* check below the segment? */
	 if (p->flags & FLOOD_TODO_BELOW) {
	    p->flags &= ~FLOOD_TODO_BELOW;
	    if (check_flood_line(bmp, p->y+1, p->lpos, p->rpos, src_color, color)) {
	       done = false;
	       p = FLOOD_LINE(c);
	    }
	 }

	 /* check above the segment? */
	 if (p->flags & FLOOD_TODO_ABOVE) {
	    p->flags &= ~FLOOD_TODO_ABOVE;
	    if (check_flood_line(bmp, p->y-1, p->lpos, p->rpos, src_color, color)) {
	       done = false;
	       /* special case shortcut for going backwards */
	       if ((c < al_get_bitmap_height(bmp)) && (c > 0))
		  c -= 2;
	    }
	 }
      }

   } while (!done);

   al_unlock_bitmap(bmp);
   al_restore_state(&old_state);
}
